/* setup.c
 * Linux CAN-bus device driver.
 * Written by Arnaud Westenberg email:arnaud@wanadoo.nl
 * Rewritten for new CAN queues by Pavel Pisa - OCERA team member
 * email:pisa@cmp.felk.cvut.cz
 * This software is released under the GPL-License.
 * Version lincan-0.2  9 Jul 2003
 */ 

#include "../include/can.h"
#include "../include/can_sysdep.h"
#include "../include/main.h"
#include "../include/devcommon.h"
#include "../include/setup.h"
#include "../include/finish.h"

extern int sja1000_register(struct chipspecops_t *chipspecops);
extern int sja1000p_register(struct chipspecops_t *chipspecops);
extern int i82527_register(struct chipspecops_t *chipspecops);

int init_hwspecops(struct candevice_t *candev, int *irqnum_p);
int init_device_struct(int card, int *chan_param_idx_p, int *irq_param_idx_p);
int init_chip_struct(struct candevice_t *candev, int chipnr, int irq, long baudrate);
int init_obj_struct(struct candevice_t *candev, struct chip_t *hostchip, int objnr);
int init_chipspecops(struct candevice_t *candev, int chipnr);

void *can_checked_malloc(size_t size)
{
	struct mem_addr *mem_new;
	void *address_p;
	
	address_p=kmalloc(size,GFP_KERNEL);
	if(address_p == NULL) {
		CANMSG("can_checked_malloc: out of the memory\n");
		return NULL;
	}

#ifdef DEBUG_MEM
	DEBUGMSG("can_checked_malloc: allocated %d bytes at %p, mem_head=%p\n",
			(int)size, address_p, mem_head);
#endif

	mem_new=(struct mem_addr *)kmalloc(sizeof(struct mem_addr),GFP_KERNEL);
	if (mem_new == NULL) {
		CANMSG("can_checked_malloc: memory list allocation error.\n");
		kfree(address_p);
		return NULL;
	}
	mem_new->next=mem_head;
	mem_new->address=address_p;
	mem_new->size=size;
	mem_head=mem_new;

	return address_p;
}

int can_checked_free(void *address_p)
{
	struct mem_addr **mem_pptr;
	struct mem_addr *mem_del=NULL;

#ifdef DEBUG_MEM
	DEBUGMSG("can_checked_free %p, mem_head=%p\n", address_p, mem_head);
#endif

	for(mem_pptr = &mem_head; (mem_del = *mem_pptr); mem_pptr = &mem_del->next) {
		if (mem_del->address != address_p)
			continue;
		*mem_pptr=mem_del->next;
		kfree(mem_del);
		kfree(address_p);
		return 0;
	}
	
	CANMSG("can_checked_free: address %p not found on the mem list\n", address_p);
	
	kfree(address_p);
	return -1;
}


int can_del_mem_list(void)
{
	struct mem_addr *mem;

#ifdef DEBUG_MEM
	DEBUGMSG("can_del_mem_list, mem_head=%p\n", mem_head);
#endif
	if(mem_head == NULL) {
		CANMSG("can_del_mem_list: no entries on the list - OK\n");
		return 0;
	}

	while((mem=mem_head) != NULL) {
		mem_head=mem->next;
		CANMSG("can_del_mem_list: deleting %p with size %d\n",
			mem->address, (int)mem->size);
		kfree(mem->address);
		kfree(mem);
	}
	
	return 0;
}

int can_request_io_region(unsigned long start, unsigned long n, const char *name)
{
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
	if(check_region(start,n)) return 0;
	request_region(start,n,name);
	return 1;
    #else
	return (request_region(start,n,name))?1:0;
    #endif
}

void can_release_io_region(unsigned long start, unsigned long n)
{
	release_region(start,n);
}

int can_request_mem_region(unsigned long start, unsigned long n, const char *name)
{
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
	return 1;
    #else
	return (request_mem_region(start,n,name))?1:0;
    #endif
}

void can_release_mem_region(unsigned long start, unsigned long n)
{
    #if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
	return;
    #else
	release_mem_region(start,n);
    #endif
}

/* This function shifts all base address structures acording to address
   translation between physical and virtual address mappings */
int can_base_addr_fixup(struct candevice_t *candev, unsigned long new_base)
{
	unsigned long offs;
	int i, j;
	
	offs=new_base-candev->dev_base_addr;
	candev->dev_base_addr=new_base;
	for(i=0;i<candev->nr_all_chips;i++){
		candev->chip[i]->chip_base_addr += offs;
		for(j=0;j<candev->chip[i]->max_objects;j++)
			candev->chip[i]->msgobj[j]->obj_base_addr += offs;
	}
	return 0;
}


int register_obj_struct(struct msgobj_t *obj, int minorbase)
{
	static int next_minor=0;
	int i;
	
	if(minorbase>=0)
		next_minor=minorbase;
	if(next_minor>=MAX_TOT_MSGOBJS)
		next_minor=0;
	i=next_minor;
	do{
		if(objects_p[i]==NULL){
			objects_p[i]=obj;
			obj->minor=i;
			next_minor=i+1;
			return 0;
		}
		if(++i >= MAX_TOT_MSGOBJS) i=0;
	}while(i!=next_minor);
	obj->minor=-1;
	return -1;
}


int register_chip_struct(struct chip_t *chip, int minorbase)
{
	static int next_chip_slot=0;
	int i;
	
	if(next_chip_slot>=MAX_TOT_CHIPS)
		next_chip_slot=0;
	i=next_chip_slot;
	do{
		if(chips_p[i]==NULL){
                	chips_p[i]=chip;

			next_chip_slot=i+1;
			return 0;
		}
		if(++i >= MAX_TOT_CHIPS) i=0;
	}while(i!=next_chip_slot);
	return -1;
}


/* The function init_hw_struct is used to initialize the hardware structure. */
int init_hw_struct(void)
{
	int i=0;
	int irq_param_idx=0;
	int chan_param_idx=0;

	hardware_p->nr_boards=0;
	while ( (hw[i] != NULL) & (i < MAX_HW_CARDS) ) {
		hardware_p->nr_boards++;

		if (init_device_struct(i, &chan_param_idx, &irq_param_idx)) {
			CANMSG("Error initializing candevice_t structures.\n");
			return -ENODEV;
		}
		i++;
	}

	return 0;
}

/* The function init_device_struct is used to initialize a single device 
 * structure.
 */
int init_device_struct(int card, int *chan_param_idx_p, int *irq_param_idx_p)
{
	struct candevice_t *candev;
	int ret;
	int irqnum;
	int chipnr;
	long bd;
	int irqsig=-1;
	
	candev=(struct candevice_t *)can_checked_malloc(sizeof(struct candevice_t));
	if (candev==NULL)
		return -ENOMEM;

        memset(candev, 0, sizeof(struct candevice_t));

	hardware_p->candevice[card]=candev;
	candev->candev_idx=card;

	candev=candev;

	candev->hwname=hw[card];
	candev->io_addr=io[card];
	candev->dev_base_addr=io[card];

	candev->hwspecops=(struct hwspecops_t *)can_checked_malloc(sizeof(struct hwspecops_t));
	if (candev->hwspecops==NULL)
		goto error_nomem;

	memset(candev->hwspecops, 0, sizeof(struct hwspecops_t));

	if (init_hwspecops(candev, &irqnum))
		goto error_nodev;

	if (candev->hwspecops->init_hw_data(candev))
		goto error_nodev;

	/* Alocate and initialize the chip structures */
	for (chipnr=0; chipnr < candev->nr_all_chips; chipnr++) {

		if(chipnr<irqnum)
			irqsig=irq[*irq_param_idx_p+chipnr];
		
		bd=baudrate[*chan_param_idx_p+chipnr];
		if(!bd) bd=baudrate[0];
	
		if ((ret=init_chip_struct(candev, chipnr, irqsig, bd*1000)))
			goto error_chip;
	}
	


	for (chipnr=0; chipnr < candev->nr_all_chips; chipnr++) {
		int m=minor[*chan_param_idx_p+chipnr];
		struct chip_t *chip=candev->chip[chipnr];
		int objnr;

		register_chip_struct(chip, m);
		
		for (objnr=0; objnr<chip->max_objects; objnr++) {
			register_obj_struct(chip->msgobj[objnr], m);
			if(m>=0) m++;
		}
	}

	*irq_param_idx_p += irqnum;
	*chan_param_idx_p += candev->nr_all_chips;

	return 0;

    error_nodev:
	ret=-ENODEV;
    error_chip:
	candevice_done(candev);
	goto error_both;

    error_nomem:
	ret=-ENOMEM;

    error_both:
	hardware_p->candevice[card]=NULL;
	can_checked_free(candev);
	return ret;
    	
}

/* The function init_chip_struct is used to initialize all chip_t structures
 * on one hardware board.
 */
int init_chip_struct(struct candevice_t *candev, int chipnr, int irq, long baudrate)
{
	struct chip_t *chip;
	int objnr;
	int ret;

	candev->chip[chipnr]=(struct chip_t *)can_checked_malloc(sizeof(struct chip_t));
	if ((chip=candev->chip[chipnr])==NULL)
		return -ENOMEM;

        memset(chip, 0, sizeof(struct chip_t));

	chip->write_register=candev->hwspecops->write_register;
	chip->read_register=candev->hwspecops->read_register;

	chip->chipspecops=can_checked_malloc(sizeof(struct chipspecops_t));
	if (chip->chipspecops==NULL)
		return -ENOMEM;

	chip->chip_idx=chipnr;
	chip->hostdevice=candev;
	chip->chip_irq=irq;
	chip->baudrate=baudrate;
	chip->flags=0x0;

	candev->hwspecops->init_chip_data(candev,chipnr);

	if (init_chipspecops(candev,chipnr))
		return -ENODEV;

	for (objnr=0; objnr<chip->max_objects; objnr++) {
		ret=init_obj_struct(candev, chip, objnr);
		if(ret<0) return ret;
	}

	return 0;
}


int init_obj_struct(struct candevice_t *candev, struct chip_t *hostchip, int objnr)
{
	struct canque_ends_t *qends;
	struct msgobj_t *obj;
	int ret;

	obj=(struct msgobj_t *)can_checked_malloc(sizeof(struct msgobj_t));
	hostchip->msgobj[objnr]=obj;
	if (obj == NULL) 
		return -ENOMEM;

        memset(obj, 0, sizeof(struct msgobj_t));
	obj->minor=-1;

	atomic_set(&obj->obj_used,0);
	INIT_LIST_HEAD(&obj->obj_users);
	init_timer(&obj->tx_timeout);

	qends = (struct canque_ends_t *)can_checked_malloc(sizeof(struct canque_ends_t));
	if(qends == NULL) return -ENOMEM;
	memset(qends, 0, sizeof(struct canque_ends_t));
	obj->hostchip=hostchip;
	obj->object=objnr+1;
	obj->qends=qends;
	obj->tx_qedge=NULL;
	obj->tx_slot=NULL;
	obj->obj_flags = 0x0;

	ret=canqueue_ends_init_chip(qends, hostchip, obj);
	if(ret<0) return ret;

	ret=candev->hwspecops->init_obj_data(hostchip,objnr);
	if(ret<0) return ret;
	
	return 0;
}


int init_hwspecops(struct candevice_t *candev, int *irqnum_p)
{
	const struct boardtype_t *brp;
	
	brp = boardtype_find(candev->hwname);
	
	if(!brp) {
		CANMSG("Sorry, hardware \"%s\" is currently not supported.\n",candev->hwname);
		return -EINVAL;
	}
	
	if(irqnum_p)
		*irqnum_p=brp->irqnum;
	brp->board_register(candev->hwspecops);

	return 0;
}


int init_chipspecops(struct candevice_t *candev, int chipnr)
{
	if (!strcmp(candev->chip[chipnr]->chip_type,"i82527")) {
		candev->chip[chipnr]->max_objects=15;
		i82527_register(candev->chip[chipnr]->chipspecops);
	} 
	if (!strcmp(candev->chip[chipnr]->chip_type,"sja1000")) {
		candev->chip[chipnr]->max_objects=1;
		sja1000_register(candev->chip[chipnr]->chipspecops);
	}
	if (!strcmp(candev->chip[chipnr]->chip_type,"sja1000p")) {
		candev->chip[chipnr]->max_objects=1;
		sja1000p_register(candev->chip[chipnr]->chipspecops);
	}

	return 0;
}

#ifndef CAN_WITH_RTL

int can_chip_setup_irq(struct chip_t *chip)
{
	if(chip==NULL)
		return -1;
	if(!chip->chipspecops->irq_handler)
		return 0;
			
	if (request_irq(chip->chip_irq,chip->chipspecops->irq_handler,SA_SHIRQ,DEVICE_NAME,chip))
		return -1;
	else {
		DEBUGMSG("Registered interrupt %d\n",chip->chip_irq);
		chip->flags |= CHIP_IRQ_SETUP;
	}
	return 1;
}


void can_chip_free_irq(struct chip_t *chip)
{
	if((chip->flags & CHIP_IRQ_SETUP) && (chip->chip_irq>=0)) {
		free_irq(chip->chip_irq, chip);
		chip->flags &= ~CHIP_IRQ_SETUP;
	}
}

#endif /*CAN_WITH_RTL*/
